## Random forests meta-model for high-resolution maize yield predictions based on low-resolution outputs from a global gridded crop model
## Supplementary for
## C. Folberth et al., 2018
## Spatio-temporal downscaling of gridded crop model yield estimates based on machine learning
## Submitted to Agricultural and Forest Meteorology
##
## This script trains a random forests model using global crop model runs. It performs predictions for a global test set and at the extent of Mexico.
## Produced graphs correspond to Figure 2 and 3 of the manuscript. Additional outputs are spatial maps of predictions and EPIC simulations.
##################################################################################################################################################################


## Clean workspace

rm(list=ls(all=TRUE))

## Load required libraries

require(hydroGOF)
require(h2o)
require(reshape2)
require(ggplot2)
require(Matrix)
require(data.table)
require(raster)
require(rasterVis)

## Start virtual cluster for H2O

localH2O = h2o.init(nthreads = 10, max_mem_size="20G", ice_root="./")

## Load dataframes gl.df (global low-resolution features and crop model simulation outputs)
## and mex.df (high-resolution features and crop model simulation outputs)

load("gl.df")
load("mex.df")

## Define function to produce sets of features

CreateFeatureVector <- function(num.mon = 0, incl.phu = F, incl.soil = F, 
                                incl.sd = F,  incl.gs = F, incl.GS = T, 
                                incl.cal = F, excl.sk = T, dframe = mex.df) {
  foi <- "" #  to make it possible to append the following features
  
  if (num.mon > 0) {
    mon <- num.mon
    foi <- c(foi,
             paste("TMX", c(1:mon), sep="_"),
             paste("TMN", c(1:mon), sep="_"),
             paste("PRCP", c(1:mon), sep="_"),
             paste("PET", c(1:mon), sep="_"),
             paste("GDD", c(1:mon), sep="_"),
             paste("RAD", c(1:mon), sep="_"),
             paste("WET", c(1:mon), sep="_"),
             paste("CMD", c(1:mon), sep="_"))
  }
  if (incl.phu == T) {
    foi <- c(foi, "PHU")
  }
  if (num.mon > 0 && incl.sd == T) {
    mon <- num.mon
    foi <- c(foi,
             paste("TMXsd", c(1:mon), sep="_"),
             paste("TMNsd", c(1:mon), sep="_"),
             paste("PRCPsd", c(1:mon), sep="_"),
             paste("RADsd", c(1:mon), sep="_"),
             paste("PETsd", c(1:mon), sep="_"),
             paste("GDDsd", c(1:mon), sep="_"),
             paste("WETsd", c(1:mon), sep="_"),
             paste("CMDsd", c(1:mon), sep="_"))
  }
  if (incl.soil == T) {
    foi <- c(foi,
             "SLP", "DEPTH", "SAND", "CLAY", 
             "PAW", "LVP", "HG", "PH", "CEC",
             "BD", "SB", "CARB", "ROK", "OC")
  }
  if (incl.GS == T) {
    foi <- c(foi, names(dframe)[grep('GS', names(dframe))])
  }
  if (incl.gs == T) {
    foi <- c(foi, names(dframe)[grep('gs', names(dframe))])
  }
  if (incl.cal == T) {
    foi <- c(foi, names(dframe)[grep('cal', names(dframe))])
  }
  if ( (excl.sk == T) & (incl.GS == T) ) {
    foi <- foi[- grep("sk", foi)]
  }
  foi <- foi[grep("GSET", foi, invert=T)]
  foi <- foi[!foi == ""]
  if (length(foi) == 0) {
    print("Features are not defined!!!")
  }
  return(foi)
}

## Split train and test data by ratio 75/25

samp.size <- floor(0.75 * nrow(gl.df))
set.seed(123)
gl.spl <- sample(seq_len(nrow(gl.df)), size = samp.size)

## Set vector for water management scenarios

irscen_vec <- c("RF", "IR") # RF=rainfed; IR=irrigated

## Select water management (here rainfed)

j <- 1

## Set water suply scenario and identifier for output files

irscen <- irscen_vec[j]
if (irscen == "RF") { irscen_name <- "rainfed" }
if (irscen == "IR") { irscen_name <- "irrigated" }

## Set target variable. Here crop yield (YLDG). Other options are growing season evapotranspiration (GSET) and crop available water (CAW) 

target <- paste0("YLDG_", irscen)

## Construct dataframe of feature subsets as defined in Table 2 of the manuscript

feature_sets <- c("monthly climate", "annual climate", "growing season climate", "complete climate")
features.df <- data.frame(feature_sets=feature_sets, nummo=c(6,0,0,6), phu=as.logical(c("T", "T", "T", "T")), soil=as.logical(c("T", "T", "T", "T")), sd=as.logical(c("F", "F", "F", "T")),
							gs=as.logical(c("F", "F", "F", "T")), GS=as.logical(c("F", "F", "T", "T")), calYR=as.logical(c("F", "T", "F", "T")), sk=as.logical(c("T", "T", "T", "F")))

## Select feature subset (here monthly climate)

k <- 1

## Generate list of features

features.fit <- CreateFeatureVector(num.mon = features.df$nummo[k], incl.phu = features.df$phu[k], 
									incl.soil = (features.df$soil[k]), incl.sd = (features.df$sd[k]),
									incl.gs = (features.df$gs[k]), incl.GS = (features.df$GS[k]),
									incl.cal = (features.df$calYR[k]), excl.sk = (features.df$sk[k]), dframe = gl.df)

## Split global dataframe in train and test data

train.df <- gl.df[gl.spl,  c(target, features.fit)]
test.df <- gl.df[-gl.spl,  c(target, features.fit)]

## Convert to H2O frame

train.df.hex = as.h2o(train.df, destination_frame = "train.df.hex")
test.df.hex = as.h2o(test.df, destination_frame = "test.df.hex")

## Remove original test and training dataframes

rm(train.df)
rm(test.df)

## Create vectors of crop model estimates corresponding to training and test dataframes

gl.train = gl.df[gl.spl, target]
gl.test = gl.df[-gl.spl, target]

## Train Random Forests meta-model

gl.fit <- h2o.randomForest(y = target, x = features.fit,
				model_id="rf.model", training_frame = train.df.hex,
				validation_frame=train.df.hex, ntree = 80, seed=222,
				max_depth=30
				);

## Do predictions for training and test data

gl.predictions.train <- as.data.frame(h2o.predict(gl.fit, train.df.hex, na.action=na.pass))$predict
gl.predictions.test <- as.data.frame(h2o.predict(gl.fit, test.df.hex, na.action=na.pass))$predict

## Clean up and free memory

#rm(gl.model.train)
#rm(gl.model.test)
gc()

## Plot densities of simulated and predicted values for global test data

gl_test_sim_pred_df <- data.frame(Simulated=gl.test, Predicted=gl.predictions.test)
gl_test_sim_pred_df_m <- melt(gl_test_sim_pred_df)

gl_density <- ggplot(data=gl_test_sim_pred_df_m, geom="blank")
gl_density <- gl_density + geom_density(aes(x=value, col=variable, linetype=variable), size=1)
gl_density <- gl_density + scale_colour_manual(values=c("grey28", "dodgerblue2"), labels=c("Simulated      ", "Predicted"))
gl_density <- gl_density + scale_linetype_discrete(labels=c("Simulated      ", "Predicted"))
gl_density <- gl_density + theme_bw()
gl_density <- gl_density + xlab("Maize yield [t ha-1]")
gl_density <- gl_density + ylab("Density")
gl_density <- gl_density + theme(legend.title=element_blank())
gl_density <- gl_density + theme(legend.position="top")
gl_density <- gl_density + theme(legend.direction="horizontal")
#gl_density

ggsave(paste0("./randomforests_global_test_set_density_", features.df$feature_sets[k], "_", tolower(target), ".png"), gl_density, height=4, width=5, dpi=600)

## Plot hexbins and regression for global test data

randomforests_gl_reg_r2 <- summary(lm(gl.predictions.test~gl.test))$r.squared
randomforests_gl_reg_sl <- summary(lm(gl.predictions.test~gl.test))$coefficients[2]
randomforests_gl_reg_ic <- summary(lm(gl.predictions.test~gl.test))$coefficients[1]
randomforests_gl_reg_mae <- mae(gl.predictions.test, gl.test)

randomforests_gl_reg_eq <- data.frame(x=c(1,1,1),
			y=c(17,18,19),
			eq=c(paste("y=", sprintf("%.2f", randomforests_gl_reg_sl), "x+", sprintf("%.2f", randomforests_gl_reg_ic), sep=""),
			paste("R2=", sprintf("%.2f", randomforests_gl_reg_r2), sep=""),			
			paste("MAE=", sprintf("%.2f", randomforests_gl_reg_mae), sep="")))

gl_hexbin <- ggplot(data=gl_test_sim_pred_df, geom="blank")
gl_hexbin <- gl_hexbin + theme_bw()
gl_hexbin <- gl_hexbin + geom_hex(data=gl_test_sim_pred_df, aes(x=Simulated, y=Predicted))
gl_hexbin <- gl_hexbin + geom_abline(intercept=0, slope=1, col="red", linetype=2)
gl_hexbin <- gl_hexbin + geom_abline(intercept=randomforests_gl_reg_ic, slope=randomforests_gl_reg_sl, col="darkgrey")
gl_hexbin <- gl_hexbin + geom_text(data=randomforests_gl_reg_eq, aes(x=x, y=y, label=eq), hjust=0)
gl_hexbin <- gl_hexbin + theme(legend.position="top")
gl_hexbin <- gl_hexbin + theme(legend.direction="horizontal")
gl_hexbin <- gl_hexbin + guides(fill = guide_colourbar(barwidth=15))
#gl_hexbin

ggsave(paste0("./randomforests_global_test_set_hexbin_", features.df$feature_sets[k], "_", tolower(target), ".png"), gl_hexbin, height=5, width=5, dpi=600)

## Extract variable importance determined by xgboost model, plot, and store table

# Extract variable importance
varimp <- as.data.frame(h2o.varimp(gl.fit))
varimp_df <- as.data.frame(varimp[c(1:20), c(1, 4)])
names(varimp_df) <- c("Feature", "Gain")
varimp_df$FeatureF <- factor(varimp_df$Feature, levels=rev(varimp_df$Feature))

# Set colors by variable domain
varimp_df$FillCol <- "burlywood4"
varimp_df$Type <- "Site"

var_vec <- c("TMX", "TMN", "PRCP", "PET", "RAD", "WET", "CMD", "GDD", "PHU", "LVP")
FillCol_vec <- c("darkred", "darkred", "cyan3", "orangered", "darkgoldenrod1", "cyan3", "darkorchid1", "darkred", "darkkhaki", "darkkhaki")
type_vec <- c("Temperature", "Temperature", "Precipitation", "PET", "Radiation", "Precipitation", "CMD", "Temperature", "Cultivar", "Cultivar")

for (i in 1:length(FillCol_vec))
	{
	varimp_df$FillCol[(grep(var_vec[i], varimp_df$Feature))] <- FillCol_vec[i]
	varimp_df$Type[(grep(var_vec[i], varimp_df$Feature))] <- type_vec[i]
	}

# Produce plot
gl_varimp <- ggplot(varimp_df, geom="blank", na.rm=T)
gl_varimp <- gl_varimp + geom_point(data=varimp_df, aes(x=(FeatureF), y=(Gain*100)), col=(varimp_df$FillCol), na.rm=T, size=5)
gl_varimp <- gl_varimp + scale_colour_manual(breaks=varimp_df$Feature, values=varimp_df$FillCol)
gl_varimp <- gl_varimp + theme_bw()
gl_varimp <- gl_varimp + ylab("Relative importance [%]")
gl_varimp <- gl_varimp + xlab("")
gl_varimp <- gl_varimp + coord_flip()
gl_varimp <- gl_varimp + scale_y_continuous(trans="log1p")
gl_varimp <- gl_varimp + theme(legend.title=element_blank())
gl_varimp <- gl_varimp + theme(legend.position="none")
#gl_varimp

ggsave(paste0("./randomforests_global_varimp_", features.df$feature_sets[k], "_", tolower(target), ".png"), gl_varimp, height=5, width=5, dpi=500)



### Do evaluations for Mexico

## Convert dataframe for Mexico to sparse model matrix for xgboost predictions

mex.Y = mex.df[, target]
mex.df.hex <- as.h2o(mex.df[, c(target, features.fit)])
mex.predictions <- as.data.frame(h2o.predict(gl.fit, mex.df.hex, na.action=na.pass))$predict
rm(mex.df.hex)

## Plot densities of simulated and predicted values for Mexico

mex_sim_pred_df <- data.frame(Simulated=mex.Y, Predicted=mex.predictions)
mex_sim_pred_df_m <- melt(mex_sim_pred_df)

mex_density <- ggplot(data=mex_sim_pred_df_m, geom="blank")
mex_density <- mex_density + geom_density(aes(x=value, col=variable, linetype=variable), size=1)
mex_density <- mex_density + scale_colour_manual(values=c("grey28", "dodgerblue2"), labels=c("Simulated      ", "Predicted"))
mex_density <- mex_density + scale_linetype_discrete(labels=c("Simulated      ", "Predicted"))
mex_density <- mex_density + theme_bw()
mex_density <- mex_density + xlab("Maize yield [t ha-1]")
mex_density <- mex_density + ylab("Density")
mex_density <- mex_density + theme(legend.title=element_blank())
mex_density <- mex_density + theme(legend.position="top")
mex_density <- mex_density + theme(legend.direction="horizontal")
#mex_density

ggsave(paste0("./randomforests_mexico_density_", features.df$feature_sets[k], "_", tolower(target), ".png"), mex_density, height=4, width=5, dpi=600)

## Plot hexbins and regression for Mexico

randomforests_mex_reg_r2 <- summary(lm(mex.predictions~mex.Y))$r.squared
randomforests_mex_reg_sl <- summary(lm(mex.predictions~mex.Y))$coefficients[2]
randomforests_mex_reg_ic <- summary(lm(mex.predictions~mex.Y))$coefficients[1]
randomforests_mex_reg_mae <- mae(mex.predictions, mex.Y)

randomforests_mex_reg_eq <- data.frame(x=c(1,1,1),
			y=c(17,18,19),
			eq=c(paste("y=", sprintf("%.2f", randomforests_mex_reg_sl), "x+", sprintf("%.2f", randomforests_mex_reg_ic), sep=""),
			paste("R2=", sprintf("%.2f", randomforests_mex_reg_r2), sep=""),			
			paste("MAE=", sprintf("%.2f", randomforests_mex_reg_mae), sep="")))

mex_hexbin <- ggplot(data=mex_sim_pred_df, geom="blank")
mex_hexbin <- mex_hexbin + theme_bw()
mex_hexbin <- mex_hexbin + geom_hex(data=mex_sim_pred_df, aes(x=Simulated, y=Predicted))
mex_hexbin <- mex_hexbin + geom_abline(intercept=0, slope=1, col="red", linetype=2)
mex_hexbin <- mex_hexbin + geom_abline(intercept=randomforests_mex_reg_ic, slope=randomforests_mex_reg_sl, col="darkgrey")
mex_hexbin <- mex_hexbin + geom_text(data=randomforests_mex_reg_eq, aes(x=x, y=y, label=eq), hjust=0)
mex_hexbin <- mex_hexbin + theme(legend.position="top")
mex_hexbin <- mex_hexbin + theme(legend.direction="horizontal")
mex_hexbin <- mex_hexbin + guides(fill = guide_colourbar(barwidth=15))
#mex_hexbin

ggsave(paste0("./randomforests_mexico_hexbin_", features.df$feature_sets[k], "_", tolower(target), ".png"), mex_hexbin, height=5, width=5, dpi=600)



### Produce raster datasets for predictions and EPIC simulations for Mexico

# Load table of coordinates and simulation units at 1k resolution for Mexico
load("simu1k_df")

# Selected year (here 1990)
yr_sel <- 1990

# Define mapping theme
mapTheme <- rasterTheme(region=brewer.pal(9,"YlGnBu"))
mapTheme$fontsize$text <- 50

# Add predictions to mex.df data frame
mex.df$predictions <- mex.predictions

# Merge coordinates and EPIC simulated values for selected year
coord_sim_df <- simu1k_df[, c("x", "y")]
simulated_vals_vec <- mex.df[mex.df$YEAR==yr_sel, target]

coord_sim_df$simulated <- simulated_vals_vec[(match(simu1k_df$SimUID, mex.df$SimUID[(mex.df$YEAR==yr_sel)]))]
simulated_vals_ras <- rasterFromXYZ(coord_sim_df[, c("x", "y", "simulated")])
values(simulated_vals_ras)[values(simulated_vals_ras)==-9999] <- NA
plt <- levelplot(simulated_vals_ras, margin=F, par.settings=mapTheme, xlab=NULL, ylab=NULL, maxpixels=1e+06)
png(paste0("./randomforests_mexico_map_", tolower(target), "_sim_", yr_sel, ".png"), width = 2400, height = 1800, units = "px", pointsize = 24)
print(plt)
dev.off()

# Merge coordinates and predicted values for selected year
coord_pred_df <- simu1k_df[, c("x", "y")]
predictions_vals_vec <- mex.df[mex.df$YEAR==yr_sel, "predictions"]

coord_pred_df$predictions <- predictions_vals_vec[(match(simu1k_df$SimUID, mex.df$SimUID[(mex.df$YEAR==yr_sel)]))]
predictions_vals_ras <- rasterFromXYZ(coord_pred_df[, c("x", "y", "predictions")])
values(predictions_vals_ras)[values(predictions_vals_ras)==-9999] <- NA
plt <- levelplot(predictions_vals_ras, margin=F, par.settings=mapTheme, xlab=NULL, ylab=NULL, maxpixels=1e+06)
png(paste0("./randomforests_mexico_map_", tolower(target), "_pred_", yr_sel, ".png"), width = 2400, height = 1800, units = "px", pointsize = 24)
print(plt)
dev.off()

## Shutdown virtual H2O cluster

h2o.shutdown()

